# Forecasting/autoforecast/src/models/keras_models.py
import numpy as np
from keras import Model
from keras.layers import Dense, LSTM, Embedding, Input
from sklearn.preprocessing import MinMaxScaler
import tensorflow as tf


class BaseKeras():
    def __init__(self, train, n_input=12, n_features=1):
        self.scaler = MinMaxScaler()
        self.scaler.fit(train.target.values.reshape(-1, 1))
        train['target_scaled'] = self.scaler.transform(train.target.values.reshape(-1, 1))
        self.y_train = train['target_scaled'].values
        self.n_input = n_input
        self.n_features = 1
        self.X_train = np.array([self.y_train[i:i+self.n_input] for i in range(len(self.y_train)-self.n_input)])
        self.y_train = np.array([self.y_train[i+1:i+self.n_input+1] for i in range(len(self.y_train)-self.n_input)])

    def fit(self, *args):
        self.model = self.keras_model(self.n_input, self.n_features)
        self.model.fit(self.X_train, self.y_train, epochs=10, validation_split=0.1, verbose=0)

    def predict(self, *args):
        pred_list = self.predict_by_batch(
            self.model, self.X_train, self.n_input, self.n_features, self.scaler)
        return pred_list

    @staticmethod
    def keras_model(n_input, n_features, hidden_size_dense=3):
        inputs = Input(shape=(n_input, n_features))
        x = Dense(hidden_size_dense, activation='relu')(inputs)
        outputs = Dense(units=1)(x)
        model = Model(inputs=inputs, outputs=outputs)
        loss = tf.keras.losses.MeanAbsolutePercentageError()
        optimizer = tf.keras.optimizers.Adam()
        model.compile(
            loss=loss,
            optimizer=optimizer
        )
        return model

    @staticmethod
    def predict_by_batch(model, X_train, n_input, n_features, scaler):
        pred_list = []
        batch = X_train[-1].reshape(-1, 1)
        for i in range(n_input):  
            pred_list.append(model.predict(batch)[0]) 
            batch = np.append(batch[1:], pred_list[i].reshape(-1, 1), axis=0)
        pred_list = np.concatenate([i for i in pred_list]).reshape(-1, 1)
        return np.concatenate(scaler.inverse_transform(pred_list))


class LSTMKeras(BaseKeras):
    def __init__(self, train):
        return super().__init__(train=train)

    def fit(self, *args):
        return super().fit(*args)

    def predict(self, *args):
        return super().predict(*args)

    @staticmethod
    def keras_model(n_input, n_features, hidden_size_lstm=3):
        inputs = Input(shape=(n_input, n_features))
        x = LSTM(hidden_size_lstm, activation='relu')(inputs)
        outputs = Dense(units=1)(x)
        model = Model(inputs=inputs, outputs=outputs)
        loss = tf.keras.losses.MeanAbsolutePercentageError()
        optimizer = tf.keras.optimizers.Adam()
        model.compile(
            loss=loss,
            optimizer=optimizer
        )
        return model



import numpy as np
from sklearn.metrics import (
    mean_absolute_error, mean_absolute_percentage_error, mean_squared_error
)


def encode(data, col='bank'):
    map_col_to_col_id = {col: col_id for col_id, col in enumerate(data[col].unique())}
    data[f'{col}_token'] = data[col].map(map_col_to_col_id)
    return data, map_col_to_col_id


def smape_score(y_test, y_pred):
    """
    https://www.statology.org/smape-python/
    """
    y_test_ = []
    y_pred_ = []
    for test, pred in zip(y_test, y_pred):
        if abs(test) < 50 and abs(pred) < 50:
            y_test_.append(1)
            y_pred_.append(1)
        elif abs(test - pred) < 100:
            y_test_.append(1)
            y_pred_.append(1)
        else:
            y_test_.append(test)
            y_pred_.append(pred)

    y_test = np.array(y_test_)
    y_pred = np.array(y_pred_)
    if len(y_test) == 0:
        return 0.0
    return 1/len(y_test) * np.sum(2 * np.abs(y_pred-y_test) / (np.abs(y_test) + np.abs(y_pred))*100)


from sklearn.metrics import mean_absolute_percentage_error, mean_absolute_error, mean_squared_error


def get_metrics(y_pred, y_test, y_naive=None):
    mse = mean_squared_error(y_pred, y_test)
    rmse = np.sqrt(mse)
    rmsle = np.log(rmse)
    mape = mean_absolute_percentage_error(y_test, y_pred)
    mae = mean_absolute_error(y_true=y_test, y_pred=y_pred)
    y_pred = np.array(y_pred)
    y_test = np.array(y_test)
    # https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error#
    smape = smape_score(y_pred, y_test)
    y_pred = np.array([0.0 if pred_ < 1e-6 else pred_ for pred_ in y_pred])
    mase = None
    if y_naive is not None:
        mae_naive = mean_absolute_error(y_true=y_test, y_pred=y_naive)
        if mae_naive == 0.0:
            if mae != 0.0:
                mae_naive = mae
            else:
                mase = 0.0
        mae_naive = mae if mae_naive == 0.0 else mae_naive
        if mase is None:
            mase = mae / mae_naive
    return {'mse': mse, 'rmse': rmse, 'rmsle': rmsle, 'mape': mape,
            'mae': mae, 'smape': smape, 'mase': mase}


# Forecasting/autoforecast/main.py
import numpy as np
import pandas as pd
import time 
import math

from src.utils.utils import *
from src.models.models import get_dict_models


class AutoForecast():
    def __init__(self, train):
        self.dict_models = get_dict_models(train)

    # custom AutoForecast
    def run_auto_forecast(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        X_test: np.ndarray,
        y_test: np.ndarray,
        verbose: bool = False,
        max_time_in_sec: int = 360
    ):
        """
        Args:
            :X_train: np.ndarray
            :y_train: np.ndarray
            :X_test: np.ndarray
            :y_test: np.ndarray
            :verbose: bool = False
            :max_time_in_sec: int = 360
        """
        print('AutoForecast starting...')
        start = time.time()
        dict_metrics = {}
        dict_pred = {}
        for model_str, model in self.dict_models.items():

            model = model
            model.fit(X_train, y_train)
            y_pred = model.predict(X_test)
            dict_pred[model_str] = y_pred
            if verbose:
                print(model_str)
                print(f'pred={y_pred}')
            metrics = get_metrics(y_test=y_test, y_pred=y_pred)
            dict_metrics[model_str] = metrics
            if start-time.time() >= max_time_in_sec:
                return {
                    'dict_metrics': dict_metrics,
                    'dict_pred': dict_pred
                }
        # print best models
        # RMSE
        print('Best models according to RMSE metrics:')
        dict_metrics_rmse = {k: v['rmse'] for k, v in dict_metrics.items()}
        # sorted dict
        dict_metrics_rmse = dict(sorted(dict_metrics_rmse.items(), key=lambda item: item[1]))
        print(dict_metrics_rmse)
        print('Best models according to MAPE metrics:')
        # MAPE
        dict_metrics_mape = {k: v['mape'] for k, v in dict_metrics.items()}
        # sorted dict
        dict_metrics_mape = dict(sorted(dict_metrics_mape.items(), key=lambda item: item[1]))
        print(dict_metrics_mape)
        print(f'AutoForecast done in {round(time.time()-start, 2)}s.')
        return {
            'dict_metrics': dict_metrics,
            'dict_pred': dict_pred
        }


def main():
    # settings
    # lists of features name
    list_cat_feat = ['timestamp']
    # lists of features name tokenized
    list_num_feat = []

    from src.data.import_bitcoin_price import get_price_for_last_n_days
    from src.models.models_baseline import BaselineMean
    from src.utils.metrics import get_metrics

    df_price = get_price_for_last_n_days(n=120, type='spot', currency_pair='BTC-USD')
    df_price = df_price.rename(columns={'price': 'target'})
    print(df_price)

    ind_cutoff = int(df_price.shape[0]* 0.8)
    ind_cutoff = 12
    train = df_price[:-ind_cutoff]
    test = df_price[-ind_cutoff:]
    print(train.shape, test.shape)

    cols = list_cat_feat + list_num_feat
    X_train = train[cols].values
    y_train = train['target'].values
    X_test = test[cols].values
    y_test = test['target'].values

    # main.py
    res_auto_forecast = AutoForecast(train).run_auto_forecast(
        X_train, y_train, X_test, y_test,
        verbose=True, max_time_in_sec=600
    )

    dict_metrics = res_auto_forecast['dict_metrics']
    dict_pred = res_auto_forecast['dict_pred']


if __name__ == '__main__':
    main()



# Forecasting/autoforecast/src/models/baseline_models.py
import numpy as  np


class BaselineLastYear():
    def __init__(self, y_train=None):
        self.y_train = y_train

    def fit(self, X_train, y_train):
        self.y_train = y_train

    def predict(self, X_test):
        return np.array(self.y_train[-12:] * len(X_test))


class BaselineLastValue():
    def __init__(self, y_train=None):
        self.y_train = y_train

    def fit(self, X_train, y_train):
        self.y_train = y_train

    def predict(self, X_test):
        return np.array([self.y_train[-1]] * len(X_test))

    
class BaselineMean():
    def __init__(self, y_train=None):
        self.y_train = y_train

    def fit(self, X_train, y_train):
        self.y_train = y_train

    def predict(self, X_test):
        return np.array([np.mean(self.y_train)] * len(X_test))


class BaselineMedian():
    def __init__(self, y_train=None):
        self.y_train = y_train

    def fit(self, X_train, y_train):
        self.y_train = y_train

    def predict(self, X_test):
        return np.array([np.median(self.y_train)] * len(X_test))



"""
utils_coinbase.py
"""
import sys
from typing import List
import time
import asyncio
from datetime import datetime, timedelta

import pandas as pd
from aiohttp import ClientSession
import requests


async def _get_current_price(type='buy', currency_pair='BTC-USD', date=None, session=None):
    if date is None:
        date = datetime.now().strftime('%Y-%m-%d')
    available_type = ['buy', 'sell', 'spot']
    BASE_URL = 'https://api.coinbase.com/v2/prices/'
    if currency_pair is not None:
        url = '{}{}/'.format(BASE_URL, currency_pair)
    else:
        print('Error: currency_pair is None')
        sys.exit(1)
    if type in available_type:
        url = '{}{}'.format(url, type)
    else:
        print('Error: type not available, try {}'.format(available_type))
        sys.exit(1)
    if date is not None:
        url = '{}?date={}'.format(url, date)
    try:
        async with session.get(url) as response:
            if response.status != 200:
                print('status_code {} for url {}'.format(response.status, url))
            else:
                json = response.json()
                return await json
    except Exception as e:
        print('Not able to get {}, {}'.format(url, e))
        return {}


def _get_current_price_loop(type='buy', currency_pair='BTC-USD', date=None, session=None):
    if date is None:
        date = datetime.now().strftime('%Y-%m-%d')
    available_type = ['buy', 'sell', 'spot']
    BASE_URL = 'https://api.coinbase.com/v2/prices/'
    if currency_pair is not None:
        url = '{}{}/'.format(BASE_URL, currency_pair)
    else:
        print('Error: currency_pair is None')
        sys.exit(1)
    if type in available_type:
        url = '{}{}'.format(url, type)
    else:
        print('Error: type not available, try {}'.format(available_type))
        sys.exit(1)
    if date is not None:
        url = '{}?date={}'.format(url, date)
    try:
        response = requests.get(url)
        if response.status_code != 200:
            print('status_code {} for url {}'.format(response.status_code, url))
        else:
            json = response.json()
            return json
    except Exception as e:
        print('Not able to get {}, {}'.format(url, e))
        return {}


async def future_price(dates: List[str], type, currency_pair) -> object:
    """
    Async function to do API call faster.
    input:
            :mdls: list, models we want ratings
            :token: str, openvoice token to query the API
            :endpoint: str, endpoint from OPV API
    output:
            asyncio object
    """
    tasks = []
    async with ClientSession() as session:
        for date in dates:
            task = asyncio.ensure_future(_get_current_price(type=type,
                                                            currency_pair=currency_pair,
                                                            date=date,
                                                            session=session))
            tasks.append(task)

        return await asyncio.gather(*tasks)


def get_price_for_last_n_days(n=1, type='spot', currency_pair='BTC-USD'):
    print('Fetching {} ({}) data for the last {} days...'.format(type,
                                                                 currency_pair,
                                                                 n))
    start = time.time()
    dates = []
    prices = []
    jsons = []
    #loop = asyncio.get_event_loop()
    for i in range(n, -1, -1):
        date = (datetime.now() - timedelta(days=i)).strftime('%Y-%m-%d')
        dates.append(date)
        jsons.append(_get_current_price_loop(type=type,
                                             currency_pair=currency_pair,
                                             date=date))
    """
    future = asyncio.ensure_future(future_price(dates=dates,
                                                type=type,
                                                currency_pair=currency_pair))
    jsons = loop.run_until_complete(future)
    loop.close()
    """
    prices = [float(json.get('data', {'amount': 0})['amount'])
              if json else 0
              for json in jsons]

    data = pd.DataFrame({'date': dates,
                         'price': prices,
                         'timestamp': [i+1 for i in range(len(dates))]})
    print('Data fetched in {:.2f} s'.format(time.time() - start))
    return data


# src/utils/utils.py
import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error


def encode(data, col='bank'):
    map_col_to_col_id = {col: col_id for col_id, col in enumerate(data[col].unique())}
    data[f'{col}_token'] = data[col].map(map_col_to_col_id)
    return data, map_col_to_col_id


def mean_absolute_percentage_error(y_true, y_pred): 
    return np.mean(np.abs((y_true - y_pred) / y_true)) * 100


def get_metrics(y_pred, y_test):
    mse = mean_squared_error(y_pred, y_test)
    rmse = np.sqrt(mse)
    rmsle = np.log(rmse)
    mape = mean_absolute_percentage_error(y_test, y_pred)
    mae = mean_absolute_error(y_test, y_pred)
    return {'mse': mse, 'rmse': rmse, 'rmsle': rmsle, 'mape': mape, 'mae': mae}


def split_n_last_timestep(df, n=12):
    train, test = df.iloc[:-n], df.iloc[-n:]
    return train, test


def scale_transform(train, test=None):
    scaler = MinMaxScaler()
    scaler.fit(train)
    train = scaler.transform(train)
    if test is not None:
        test = scaler.transform(test)
    return train, test, scaler


def prepare_for_training(df, cols_to_drop=['heading_id', 'company_id']):
    df = df.drop(cols_to_drop, axis=1)
    df.date = pd.to_datetime(df.date)
    df = df.set_index("date")
    if 'month' not in cols_to_drop:
        df['month'] = df['month'].astype('float64')
    return df


def train_test_split(df):
    df_ = df.copy()
    train, test = split_n_last_timestep(df, n=12)
    return train, test


def plot_pred(pred_list, df, future_dates=None):
    if future_dates is None:
        df_predict = pd.DataFrame(pred_list,
                                  index=df[-N_INPUT:].index, columns=['Prediction'])
    else:
        df_predict = pd.DataFrame(pred_list,
                                  index=future_dates[-N_INPUT:].index, columns=['Prediction'])

    df_test = pd.concat([df,df_predict], axis=1)
    plt.figure(figsize=(20, 5))
    plt.plot(df_test.index, df_test['target'])
    plt.plot(df_test.index, df_test['Prediction'], color='r')
    plt.legend(loc='best', fontsize='xx-large')
    plt.xticks(fontsize=18)
    plt.yticks(fontsize=16)
    plt.show()



import plotly.graph_objects as go
    
    y_train = (
        train_2016_.target.tolist() + train_2017_.target.tolist() 
        + train_2018_.target.tolist() + train_2019_.target.tolist()
    )
    x_2019 = list(map(str,range(201901, 201913)))

    fig = go.Figure(data=go.Scatter(x=x_2019, y=train_2019_['target'], name='2019'))
    fig.add_trace(go.Scatter(x=x_2019, y=train_2018_['target'], name='2018'))
    fig.add_trace(go.Scatter(x=x_2019, y=train_2017_['target'], name='2017'))
    fig.add_trace(go.Scatter(x=x_2019, y=train_2016_['target'], name='2016'))
    fig.add_trace(go.Scatter(x=x_2019, y=pred, name='pred'))
    title = (
        """" my title """"
        
    )
    print(title)
    fig.show()



from autoforecast.src.data.import_bitcoin_price import get_price_for_last_n_days
from autoforecast.src.models.models_baseline import BaselineMean
from autoforecast.src.utils.metrics import get_metrics

df_price = get_price_for_last_n_days(n=30, type='spot', currency_pair='BTC-USD')
print(df_price)

ind_cutoff = int(df_price.shape[0]* 0.8)
train = df_price[:ind_cutoff]
test = df_price[ind_cutoff:]

X_train = train['timestamp']
y_train = train.price
X_test = test['timestamp']
y_test = test.price

model = BaselineMean()
model.fit(X_train=X_train, y_train=y_train)

y_pred = model.predict(X_test=X_test)

metrics = get_metrics(y_test=y_test, y_pred=y_pred)
print(metrics)



from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor, AdaBoostRegressor
from sklearn.linear_model import LinearRegression, Ridge, SGDRegressor
from sklearn.svm import LinearSVR, SVR, NuSVR
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.neighbors import KNeighborsRegressor
import xgboost as xgb

from src.models.models_keras import *
from src.models.models_baseline import *
from src.models.models_time_series import *


def get_dict_models(train):
    dict_models = {
        'LSTMKeras': LSTMKeras(train=train),
        'BaseKeras': BaseKeras(train=train),
        'XGBRegressor': xgb.XGBRegressor(),
        'RandomForestRegressor': RandomForestRegressor(),
        'GradientBoostingRegressor': GradientBoostingRegressor(),
        'AdaBoostRegressor': AdaBoostRegressor(),
        'LinearRegression': LinearRegression(),
        'Ridge': Ridge(),
        'SGDRegressor': SGDRegressor(),
        'LinearSVR': LinearSVR(),
        'SVR': SVR(),
        'NuSVR': NuSVR(),
        'DecisionTreeRegressor': DecisionTreeRegressor(),
        'ExtraTreeRegressor': ExtraTreeRegressor(),
        'MLPRegressor': MLPRegressor(),
        'KNeighborsRegressor': KNeighborsRegressor(),
        'ARMA': ARMA(),
        'BaselineLastYear': BaselineLastYear(),
        'BaselineLastValue': BaselineLastValue(),
        'BaselineMean': BaselineMean(),
        'BaselineMedian': BaselineMedian(),
        'Prophet': Prophet(train=train),
    }
    return dict_models



# Forecasting/autoforecast/src/models/time_series_models.py
import numpy as np
import pandas as pd
import fbprophet
import statsmodels.api as sm


class ARMA():
    def __init__(self, order=None, period=None):
        self.order = order

    def fit(self, X_train, y_train):
        self.y_train = y_train
        aic_matrix = self.build_aic(
            y_train=self.y_train, p_max=6, q_max=6, p_min=0, q_min=0
        )
        self.order = self.best_order(aic_matrix)

    def predict(self, *arg):
        y_pred = self.predict_arma(y_train=self.y_train , order=self.order)
        return y_pred
    
    @staticmethod
    def build_aic(y_train: np.ndarray, p_max=6, q_max=6, p_min=0, q_min=0):
        aic_full = pd.DataFrame(np.zeros((6,6), dtype=float))
        # Iterate over all ARMA(p,q) models with p,q in [0,6]
        for p in range(6):
            for q in range(6):
                if p == 0 and q == 0:
                    continue
                # Estimate the model with no missing datapoints
                mod = sm.tsa.statespace.SARIMAX(y_train, order=(p,0,q), enforce_invertibility=False)
                try:
                    res = mod.fit(disp=False)
                    aic_full.iloc[p,q] = res.aic
                except:
                    aic_full.iloc[p,q] = np.nan
        return aic_full

    @staticmethod
    def best_order(aic_full: pd.DataFrame):
        aic_full.iloc[0,0]= 10.0**99
        # min column
        min_col_name = aic_full.min().idxmin()
        # min column index if needed
        min_col_idx = aic_full.columns.get_loc(min_col_name)
        # min row index
        min_row_idx = aic_full[min_col_name].idxmin()
        return (min_row_idx, 0, min_col_idx)

    @staticmethod
    def predict_arma(y_train: np.ndarray, order=(1, 0, 1)):
        mod = sm.tsa.statespace.SARIMAX(y_train, order=order)
        res = mod.fit(disp=False)
        # In-sample one-step-ahead predictions, and out-of-sample forecasts
        nforecast = 12
        predict = res.get_prediction(end=mod.nobs + nforecast)
        y_pred = predict.predicted_mean[-nforecast:]
        return np.array(y_pred)


class Prophet():
    def __init__(self, train, n_input=12):
        self.train = train
        self.model = None
        self.period = n_input

    def fit(self, X_train, y_train):
        if 'date' not in self.train.columns:
            raise "train df should have a 'date' column"
        df = pd.DataFrame({'ds': self.train.date, 'y': y_train})
        self.model = fbprophet.Prophet()
        self.model.fit(df)

    def predict(self, X_test):
        future = self.model.make_future_dataframe(periods=self.period, freq='M')
        forecast = self.model.predict(future)
        y_pred = forecast.yhat[-self.period:]
        return y_pred.values
